#
# This file contains class __menu for Pasta-Man
#

from os.path import join as jPath, exists as there
from pathlib import Path
from tkinter import *
import ttkthemes
import threading
import sys

import ttkthemes.themed_style

class __menu__:
    def __init__(self, menuparent: Menu):
        self.parent = menuparent
        
        # set style -> theme -> default: arc
        if not there(jPath(str(Path.home()), '.pastaman', '.defaulttheme')):
            self.style:ttkthemes.themed_style.ThemedStyle = ttkthemes.themed_style.ThemedStyle(theme='arc')
        else:
            with open(jPath(str(Path.home()), '.pastaman', '.defaulttheme'), 'r') as theme:
                self.defaultTheme = theme.read().replace('\n', '')
                self.style:ttkthemes.themed_style.ThemedStyle = ttkthemes.themed_style.ThemedStyle(theme=self.defaultTheme)
    
    def Themes(self):
        ThemeMenu = Menu(self.parent)
        
        self.parent.add_cascade(label="Themes", menu=ThemeMenu)
        
        # Themes Functions #
        def savetheme(theme:str):
            with open(jPath(str(Path.home()), '.pastaman', '.defaulttheme'), 'w') as themefile:
                themefile.write(theme)
            sys.exit(0)
            
        def changeThemeToKroc():
            self.style.theme_use('kroc')
            threading.Thread(target=savetheme, args=('kroc',)).start()
        
        def changeThemeToKeramik():
            self.style.theme_use('keramik')
            threading.Thread(target=savetheme, args=('keramik',)).start()
        
        def changeThemeToEquilux():
            self.style.theme_use('equilux')
            threading.Thread(target=savetheme, args=('equilix',)).start()
        
        def changeThemeToElegance():
            self.style.theme_use('elegance')
            threading.Thread(target=savetheme, args=('elegance',)).start()
        
        def changeThemeToClearlooks():
            self.style.theme_use('clearlooks')
            threading.Thread(target=savetheme, args=('clearlooks',)).start()
        
        def changeThemeToBreeze():
            self.style.theme_use('breeze')
            threading.Thread(target=savetheme, args=('breeze',)).start()
        
        def changeThemeToBlue():
            self.style.theme_use('blue')
            threading.Thread(target=savetheme, args=('blue',)).start()
        
        def changeThemeToBlack():
            self.style.theme_use('black')
            threading.Thread(target=savetheme, args=('black',)).start()
        
        def changeThemeToAquativo():
            self.style.theme_use('aquativo')
            threading.Thread(target=savetheme, args=('aquativo',)).start()
        
        def changeThemeToArc():
            self.style.theme_use('arc')
            threading.Thread(target=savetheme, args=('arc',)).start()
        
        def changeThemeToAdapta():
            self.style.theme_use('adapta')
            threading.Thread(target=savetheme, args=('adapta',)).start()
        
        def changeThemeToPlastik():
            self.style.theme_use('plastik')
            threading.Thread(target=savetheme, args=('plastik',)).start()
        
        def changeThemeToRadiance():
            self.style.theme_use('radiance')
            threading.Thread(target=savetheme, args=('radiance',)).start()
        
        def changeThemeToSmog():
            self.style.theme_use('smog')
            threading.Thread(target=savetheme, args=('smog',)).start()
        
        def changeThemeToWinXPBlue():
            self.style.theme_use('winxpblue')
            threading.Thread(target=savetheme, args=('winxpblue',)).start()
        
        def changeThemeToYaru():
            self.style.theme_use('yaru')
            threading.Thread(target=savetheme, args=('yaru',)).start()
        
        # ---> themes menu
        if self.defaultTheme=='adapta':
            ThemeMenu.add_command(label='Adapta - Default', command=changeThemeToAdapta)
        else:
            ThemeMenu.add_command(label='Adapta', command=changeThemeToAdapta)

        if self.defaultTheme=='arc':
            ThemeMenu.add_command(label='Arc - Default', command=changeThemeToArc)
        else:
            ThemeMenu.add_command(label='Arc', command=changeThemeToArc)
        
        if self.defaultTheme=='aquativo':
            ThemeMenu.add_command(label='Aquativo - Default', command=changeThemeToAquativo)
        else:
            ThemeMenu.add_command(label='Aquativo', command=changeThemeToAquativo)
        
        if self.defaultTheme=='black':
            ThemeMenu.add_command(label='Black - Default', command=changeThemeToBlack)
        else:
            ThemeMenu.add_command(label='Black', command=changeThemeToBlack)
        
        if self.defaultTheme=='blue':
            ThemeMenu.add_command(label='Blue - Default', command=changeThemeToBlue)
        else:
            ThemeMenu.add_command(label='Blue', command=changeThemeToBlue)
        
        if self.defaultTheme=='breeze':
            ThemeMenu.add_command(label='Breeze - Default', command=changeThemeToBreeze)
        else:
            ThemeMenu.add_command(label='Breeze', command=changeThemeToBreeze)
        
        if self.defaultTheme=='clearlooks':
            ThemeMenu.add_command(label='Clearlooks - Default', command=changeThemeToClearlooks)
        else:
            ThemeMenu.add_command(label='Clearlooks', command=changeThemeToClearlooks)
        
        if self.defaultTheme=='elegance':
            ThemeMenu.add_command(label='Elegance - Default', command=changeThemeToElegance)
        else:
            ThemeMenu.add_command(label='Elegance', command=changeThemeToElegance)
        
        if self.defaultTheme=='equilux':
            ThemeMenu.add_command(label='Equilux - Default', command=changeThemeToEquilux)
        else:
            ThemeMenu.add_command(label='Equilux', command=changeThemeToEquilux)
        
        if self.defaultTheme=='keramik':
            ThemeMenu.add_command(label='Keramik - Default', command=changeThemeToKeramik)
        else:
            ThemeMenu.add_command(label='Keramik', command=changeThemeToKeramik)
        
        if self.defaultTheme=='kroc':
            ThemeMenu.add_command(label='Kroc - Default', command=changeThemeToKroc)
        else:
            ThemeMenu.add_command(label='Kroc', command=changeThemeToKroc)
        
        if self.defaultTheme=='plastik':
            ThemeMenu.add_command(label='Plastik - Default', command=changeThemeToPlastik)
        else:
            ThemeMenu.add_command(label='Plastik', command=changeThemeToPlastik)
        
        if self.defaultTheme=='radiance':
            ThemeMenu.add_command(label='Radiance (Ubuntu) - Default', command=changeThemeToRadiance)
        else:
            ThemeMenu.add_command(label='Radiance (Ubuntu)', command=changeThemeToRadiance)
        
        if self.defaultTheme=='smog':
            ThemeMenu.add_command(label='Smog - Default', command=changeThemeToSmog)
        else:
            ThemeMenu.add_command(label='Smog', command=changeThemeToSmog)
        
        if self.defaultTheme=='winxpblue':
            ThemeMenu.add_command(label='Win XP - Default', command=changeThemeToWinXPBlue)
        else:
            ThemeMenu.add_command(label='Win XP', command=changeThemeToWinXPBlue)
        
        if self.defaultTheme=='yaru':
            ThemeMenu.add_command(label='Yaru - Default', command=changeThemeToYaru)
        else:
            ThemeMenu.add_command(label='Yaru', command=changeThemeToYaru)


#
# This File contains the Exceptions related to PASTA-MAN only.
#

class InvalidKeyword(Exception):
    pass

class InvalidExportType(Exception):
    pass

class NoneTypeVariable(Exception):
    pass

class OptError(Exception):
    pass


#
# This file is an architecture for all operations
#

# import project specific modules.
from pasta_man.exceptions import InvalidKeyword, InvalidExportType

# import other arguments
from cryptography.fernet import Fernet
from cryptography.hazmat.primitives.hashes import SHA256
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from base64 import urlsafe_b64encode
from pathlib import Path
from os import makedirs
from os.path import join as jPath, exists as there
from datetime import datetime
# from tabulate import tabulate
import pandas as pd
from re import match

#
# The format to save the persistent passwords file ==>
# key:value|key:value|...\n
#
#

class targets:
    def __init__(self, masterpassword: bytes, mastersalt: bytes = "pastaman".encode('ascii')):
        """Initialize targets class.

        Args:
            masterpassword (bytes): master password encoded in 'ascii'
            mastersalt (bytes, optional): master salt encoded in 'ascii'. Defaults to "pastaman".encode('ascii').
        
        Description:
            masterpassword: The Password to Encrypt All. To encode in bytes, use <password-value>.encode('ascii')
            mastersalt: Salt is addition text to strengthen the password. By default it is set to 'pastaman'. To encode in bytes, use <salt-text>.encode('ascii')
        """
        # initialize data
        self.data:list[dict] = []
        self.msalt = mastersalt
        self.mpass = masterpassword
    
    def init(self):
        # create kdf object
        _kdf = PBKDF2HMAC(
            algorithm=SHA256(),
            length=32,
            iterations=480000,
            salt=self.msalt,
        )
        
        # derive password digest
        self.passwd = urlsafe_b64encode(_kdf.derive(self.mpass))
        
        # define fernet object
        self.fernet = Fernet(self.passwd)
        
        # check for pre-existing config files => for persistent memory
        # -> if not found, create
        # --> store home path
        home = str(Path.home())
        
        # if config file present, else create an init entry.
        if there(jPath(home, '.pastaman', '.passwords')):
            # open passwords file
            with open(jPath(home, '.pastaman', '.passwords'), 'rb') as p:
                content = p.read()
            # decrypt passwords file
            content = self.fernet.decrypt(content)
            
            # divide content based on lines
            content = content.decode('ascii').split('\n')
            
            # for each line add the data to self.data
            for x in content:
                if x=="":
                    continue
                # split key-value pairs
                x = x.split('|') # [k:v, k:v, ...]
                # for each key-value pair, store key and value
                # -> in data
                data = {}
                for kv in x:
                    # split keys and values
                    kv = kv.split(':')
                    data[kv[0]] = kv[1]
                
                self.data.append(data)
            
            content = None
        else:
            # create a .passwordsfile
            with open(jPath(home, ".pastaman", '.passwords'), 'w') as pfile:
                pfile.write(f"target-type:init|target:init|username:init|password:{self.fernet.encrypt('init'.encode('ascii')).decode('ascii')}|timestamp:{datetime.today().strftime('%Y-%m-%d')}\n")
            
            with open(jPath(home, ".pastaman", '.passwords'), 'rb') as p:
                content = p.read()
            
            content = self.fernet.encrypt(content)
            
            with open(jPath(home, ".pastaman", '.passwords'), 'wb') as pfile:
                pfile.write(content)
            
            content = None
        
    
    def add(self, target:str, targettype: str, username: str, password:bytes):
        """Add data in collection.

        Args:
            target (str): target is the place where this data is valid.
            targettype (str): example -> link, app, etc.
            username (str): credential
            password (bytes): password in bytes. Encoding -> 'ascii'
        """
        # set date time format
        t = datetime.today().strftime("%Y-%m-%d")
        
        # define data instance
        data = {
            "target-type":targettype,
            "target":target,
            "username":username,
            "password":self.fernet.encrypt(password),
            "timestamp":t
        }
        
        # append to self.data for safe keeping
        self.data.append(data)
        
        # create the string to be written in .passwords file
        pstring = "target-type:"+targettype+'|'+"target:"+target+'|'+"username:"+username+'|'+"password:"+self.fernet.encrypt(password).decode('ascii')+"|"+"timestamp:"+t
        
        # decrypt the passwords file
        # -> read it
        with open(jPath(str(Path.home()), ".pastaman", '.passwords'), 'rb') as pfile:
            content = pfile.read()
        
        # decrypt
        content = self.fernet.decrypt(content)
        
        # add entry
        content = content.decode('ascii').split('\n') # split by lines
        content.append(pstring)
        
        pstring = None
        
        # generate the string again
        newcontent = ""
        for x in content:
            newcontent += x + "\n"
        
        content = None
        
        newcontent = self.fernet.encrypt(newcontent.encode('ascii'))
        
        # write it
        with open(jPath(str(Path.home()), '.pastaman', '.passwords'), 'wb') as pfile:
            pfile.write(newcontent)
        
        newcontent = None
    
    def remove(self, target:str, username: str):
        """remove an entry from the collection of all entries.

        Args:
            target (str): target name.
            username (str): username.
        """
        for dictionary in self.data:
            if dictionary['target']==target and dictionary['username']==username:
                dump = self.data.pop(self.data.index(dictionary))
                dump = None
    
    def authenticate(self, username: str ,password: bytes) -> bool:
        """check if the username has the given password

        Args:
            username (str): username 
            password (bytes): password encoded in 'ascii'. To encode in ascii, use -> <password-text>.encode('ascii')

        Returns:
            bool: True if authentication passes, else False.
        """
        for dictionary in self.data:
            if dictionary['username']==username and dictionary['password']==self.fernet.encrypt(password):
                return True
        
        return False
    
    def count(self) -> dict:
        """Dynamic count of all target-types

        Returns:
            dict: returns a dictionary with target-types as keys and their counts as their values.
        
        Description:
            Example Output:
                1. {
                    "target-type_1":23,
                    "target-type_2":10,
                    ...
                }
                
                2. {} -> could be an empty dict.
        """
        counts:dict = {}
        count_target_types: int = 0
        counted_targets: list[str] = []
        for dictionary in self.data:
            if dictionary['target-type'] not in counted_targets:
                count_target_types += 1
                counted_targets.append(dictionary['target-type'])
        
        count = 0
        for targettype in counted_targets:
            for dictionary in self.data:
                if dictionary['target-type'] == targettype:
                    count += 1
            
            counts[targettype] = count
            count = 0
        
        return counts
    
    def search(self, searchkeyword: str, keywordtype: str = "target"):
        """search an entry by these keywordtypes -> ['target', 'target-type', 'username']

        Args:
            searchkeyword (str): _description_
            keywordtype (str, optional): _description_. Defaults to "target".
        """
        valid = ['target', 'target-type', 'username']
        
        if keywordtype not in valid:
            raise InvalidKeyword(f'{keywordtype} cannot be set as keywordtype.')
        
        self.__searchresult__ = None
        
        for dictionary in self.data:
            if searchkeyword in dictionary[keywordtype]:
                self.__searchresult__ = dictionary

    def targets(self):
        ts:list[str] = []
        added = []
        for dictionary in self.data:
            if dictionary['target-type']!="init" and dictionary['target-type'] not in added:
                ts.append(dictionary['target-type'])
                added.append(dictionary['target-type'])
        
        self.__target_types__ = ts
    
    def decrypt(self, password: bytes):
        self._dec_ = self.fernet.decrypt(password).decode('ascii')
    
    def export(self, exporttype: str = "csv") -> None:
        """General purpose Export. Excludes timestamp.

        Args:
            exporttype (str, optional): specify export type. Valid -> ['csv', 'xlsx']. Defaults to "csv".
        """
        valid = ['csv', 'xlsx']
        
        if exporttype not in valid:
            raise InvalidExportType(f'{exporttype} is not supported yet.')
        
        self.DataFrame = pd.DataFrame(self.data).drop('timestamp', axis=1)
        
        if exporttype == 'csv':
            self.DataFrame.to_csv(datetime.now()+"-pastaman-passwords.csv", index = False, header = False)
        elif exporttype == 'xlsx':
            self.DataFrame.to_excel(datetime.now()+"-pastaman-passwords.xlsx", index=False, header=False)
    
    # def import(self, )
        


#!/usr/bin/env python3

# version info
__version__ = "1.0"

# import project specific modules
from pasta_man.gui import pmanager
from pasta_man.encryption import Encryption
from pasta_man.exceptions import NoneTypeVariable, OptError
from pasta_man.utilities.helptext import helptext

# import libs
from tkinter import *
from tkinter import simpledialog
from termcolor import colored
from pathlib import Path
from os.path import join as jPath, exists as there
from os import makedirs
from optioner import options
import sys
import threading

def locker(masterpassword: str):
    global encb
    enc = Encryption("pastaman".encode('ascii'), masterpassword.encode('ascii'))
    enc.lock()
    encb = enc.__encryptedstring__

def unlocker(encpassword: str):
    global dencb
    denc = Encryption("pastaman".encode("ascii"), encpassword.encode('ascii'))
    denc.unlock()
    dencb = denc.__unencryptedstring__

def checkmfile(home = str(Path.home())) -> bytes:
    # --> find out if master password is defined -> .m file
    if not there(jPath(home, '.pastaman', '.m')):
        # ask dialog
        masterpassword = simpledialog.askstring("User Input", "Enter Master Password: ")
        # check if it is empty or non
        if masterpassword=='' or masterpassword==None:
            raise NoneTypeVariable('Aborted.')
        # encode it
        masterpassword = masterpassword.encode('ascii')
        enct = threading.Thread(target=locker, args=(masterpassword.decode('ascii'),))
        enct.start()
        enct.join()
        # -> store it
        with open(jPath(home, '.pastaman', '.m'), 'wb') as m:
            m.write(encb)
    else:
        # if m is present
        # -> read it
        with open(jPath(home, '.pastaman', '.m'), 'rb') as m:
            masterpassword = m.read() # this is encrypted
        
        # -> decrypt it using the Encryption class
        # denc = Encryption("pastaman".encode('ascii'), masterpassword)
        # denc.unlock()
        denct = threading.Thread(target=unlocker, args=(masterpassword.decode('ascii'),))
        denct.start()
        denct.join()
        
        # fetch it again
        masterpassword = dencb
    
    return masterpassword

def main():
    # define global vars
    global encb, dencb
    
    # define arguments
    shortargs = ['h', 'v']
    longargs = ['help', 'version']
    
    optctrl = options(shortargs, longargs, sys.argv[1:], ifthisthennotthat=[['help', 'h'],['v', 'version']])
    args, check, error, falseargs = optctrl._argparse()
    
    if not check:
        raise OptError(error)
    else:
        h = helptext()
        h.__version__ = __version__
        if '-v' in args or '--version' in args:
            h.showver()
        elif '-h' in args or '--help' in args:
            h.helper()
    
    # -> home folder
    home = str(Path.home())
    # -> find out if .pastaman folder is there
    if not there(jPath(home, '.pastaman')):
        makedirs(jPath(home, '.pastaman'))
    
    masterpassword = checkmfile()
    
    # define main root gui window
    rootwindow = Tk()
    
    # create pass manager object.
    pManager = pmanager(rootwindow, masterpassword)
    pManager._makeinitscreen_()
    
    rootwindow.mainloop()

if __name__=="__main__":
    encb:bytes = ''.encode('ascii')
    dencb:bytes = ''.encode('ascii')
    try:
        main()
    except KeyboardInterrupt:
        print("\n"+colored("KEYBOARD INTERRUPT", 'red'))
        sys.exit(1)


from cryptography.fernet import Fernet
from cryptography.hazmat.primitives.hashes import SHA256
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from base64 import urlsafe_b64encode

class Encryption:
    def __init__(self, password: bytes, string: bytes):
        self._kdf = PBKDF2HMAC(
            algorithm=SHA256(),
            length=32,
            iterations=480000,
            salt="pastaman".encode('ascii')
        )
        
        self.str = string
        
        _p = urlsafe_b64encode(self._kdf.derive(password))
        
        self._fernet = Fernet(_p)
    
    def lock(self):
        self.__encryptedstring__ = self._fernet.encrypt(self.str)
    
    def unlock(self):
        self.__unencryptedstring__ = self._fernet.decrypt(self.str)


from termcolor import colored
import sys

class helptext:
    def __init__(self):
        self.__version__ = ''
    
    def helper(self):
        print(colored('Pasta Man', 'blue'), colored(f'v{self.__version__}', 'red'))
        print("helptext")
        print("  |  -h or --help     : show this help and exit.")
        print("  |  -v or --version  : show version and exit.")
        sys.exit(0)
    
    def showver(self):
        print(colored('Pasta Man', 'blue'), colored(f'v{self.__version__}', 'red'))
        print('author: d33pster', 'GitHub:', colored('https://github.com/d33pster/pasta-man', 'light_blue'))
        sys.exit(0)


#
# This is the pmanager class file
#

# import project specific modules
from pasta_man.targets import targets
from pasta_man.encryption import Encryption
from pasta_man.utilities.pasta_menu import __menu__

# import libs
from tkinter import *
from tkinter import ttk, simpledialog, messagebox
from os.path import join as jPath
from pathlib import Path
import pyperclip
import threading

class pmanager:
    def __init__(self, master:Tk, masterpassword: bytes):
        # create a master object
        self.parent = master
        # set title
        self.parent.title('Pasta-Man')
        # set geometry
        self.parent.geometry('530x300+400+280')
        # set architecture
        self.arch = targets(masterpassword)
        
        # set menu
        menu = Menu(self.parent)
        self.parent.config(menu=menu)
        # -> add menus
        self.pasta_menu = __menu__(menuparent=menu)
        self.pasta_menu.Themes()
        
        # create Enclosing Frame
        self.EF = ttk.Frame(self.parent)
        self.EF.pack(fill=BOTH, expand=True)
        
        self.initthread = threading.Thread(target=self.arch.init).start()
    
    def _makeinitscreen_(self):
        # create two notebooks
        # -> Add and Get --> Get will have submods
        self.notebook = ttk.Notebook(self.EF)
        self.notebook.pack(expand=True, fill=BOTH)
        
        # -> create the tabs --> Add
        self.hometab = Frame(self.notebook, bg='black')
        self.addtab = ttk.Frame(self.notebook)
        self.gettab = ttk.Frame(self.notebook)
        
        # -> add the tabs under notebook
        self.notebook.add(self.hometab, text='Home')
        self.notebook.add(self.addtab, text='Add')
        self.notebook.add(self.gettab, text='Fetch')
        
        # -> invoke _makeHome)=_ to create Home screen
        self._makeHome_()
        
        # -> invoke _makeAdd_ to create Add Screen
        self._makeAdd_()
        
        # -> invoke _makeFetch_ to create Fetch Screen
        self._makeFetch_()
    
    def _makeHome_(self):
        # -> create Enclosing frame
        self.HEF = Frame(self.hometab, bg='black')
        self.HEF.pack(expand=True, fill=BOTH)
        
        # -> Create gif frames OR LABEL
        self.homelabel = Label(self.HEF, text="Pasta-Man is your very own Password Manager\nWith tripple layer security and the all new search.", font=('Verdana', 18))
        self.homelabel.place(anchor='center', relx=0.5, rely=0.35)
        
        self.homelabel.config(bg='black', fg='white')
        
        # -> create a button
        self.homeNextLabel = Label(self.HEF, text="Next >", bg='black', fg='white', font=('Verdana', 15))
        self.homeNextLabel.place(anchor='center', relx=0.5, rely=0.67)
        
        # -> bind command
        self.homeNextLabel.bind('<Button-1>', lambda e: self.homenext())
    
    def homenext(self):
        self.notebook.select(1)
        # -> set focus to targetEntry
        self.targetEntry.focus()
    
    def _makeFetch_(self):
        # -> create eclosing frame under Fetch
        self.FEF = ttk.Frame(self.gettab)
        self.FEF.pack(fill=BOTH, expand=True)
        
        # # -> find targets and fetch it later
        # t1 = threading.Thread(target=self.arch.targets)
        # t1.start()
        
        # -> create search Entry and label
        self.searchLabel = ttk.Label(self.FEF, text='Search:', font=('Verdana', 14))
        self.searchLabel.place(anchor='center', relx=0.12,rely=0.16 )
        
        # ---> Entry
        self.varsearch = StringVar()
        self.varsearch.set('hint: search keywords')
        self.searchEntry = ttk.Entry(self.FEF, textvariable=self.varsearch, width=24)
        self.searchEntry.place(anchor='center', relx=0.295, rely=0.3)
        
        # -> create search button
        self.searchbutton_var = StringVar()
        self.searchbutton_var.set('[ Search ]')
        self.searchbutton = ttk.Button(self.FEF, text='[ Search ]', default=ACTIVE, command=self.search, textvariable=self.searchbutton_var)
        self.searchbutton.place(anchor='center', relx=0.5, rely=0.65)
        
        # -> create target type label
        self.tarLabel = ttk.Label(self.FEF, text='Type:', font=('Verdana', 14))
        self.tarLabel.place(anchor='center', relx=0.67,rely=0.16)
        
        # -> create dropdown menu.
        # ----> create a var for option
        self.varoption = StringVar()
        # t1.join()
        self.targtypelist = ttk.OptionMenu(self.FEF, self.varoption, *['target', 'target-type', 'username'])
        self.targtypelist.place(anchor='center', relx=0.75, rely=0.3, relwidth=0.35)
        
        # -> create a status label
        self.fetchstatusvar = StringVar()
        self.fetchstatus = ttk.Label(self.FEF, textvariable=self.fetchstatusvar)
        self.fetchstatus.place(anchor='center', relx=0.5, rely=0.79)
    
    def search(self):
        # -> fetch value and search:
        t1 = threading.Thread(target=self.arch.search, args=(self.varsearch.get(), self.varoption.get()))
        t1.start()
        t1.join()
        
        if self.arch.__searchresult__!=None:
            self.fetchstatusvar.set('Match Found!')
            self.fetchstatus.after(5000, self.updatefetch)
        else:
            self.fetchstatusvar.set('No Match!')
            self.fetchstatus.after(4000, self.fetchstatusReset)
    
    def updatefetch(self):
        # -> forget fetch status
        self.fetchstatus.place_forget()
        self.fetchstatus.place(anchor='center', relx=0.5, rely=0.5)
        # -> rename search button and forget it
        # self.searchbutton_var.set('[ Re-Search ]')
        self.searchbutton.place_forget()
        # self.searchbutton.place(anchor='center')
        
        # -> create 3 buttons => copy to clipboard, Remove, Re-search
        self.copyToClipboardButton = ttk.Button(self.FEF, text='[ Copy To Clipboard ]', default=ACTIVE, command=self.copyToClipboard)
        self.removeButton = ttk.Button(self.FEF, text='[ Remove ]', command=self.remove)
        self.researchButton = ttk.Button(self.FEF, text='[ Re-Search ]', command=self.research)
        
        self.copyToClipboardButton.place(anchor='center', relx=0.24, rely=0.62)
        self.removeButton.place(anchor='center', relx=0.54, rely=0.62)
        self.researchButton.place(anchor='center', relx=0.8, rely=0.62)
    
    def research(self):
        # reinit the _makeFetch_ to refresh newly added Entries
        for widget in self.gettab.winfo_children():
            widget.destroy()
        
        self._makeFetch_()
    
    def remove(self):
        dump = self.arch.data.pop(self.arch.data.index(self.arch.__searchresult__))
        dump = None
        
        # reinit the _makeFetch_ to refresh newly added Entries
        for widget in self.gettab.winfo_children():
            widget.destroy()
        
        self._makeFetch_()
    
    def decryptthread(self, masterpassword: str):
        denc = Encryption("pastaman".encode('ascii'), masterpassword.encode('ascii'))
        denc.unlock()
        self.den = denc.__unencryptedstring__
    
    def copyToClipboard(self):
        
        # prepare alleged pass
        allegedpass = simpledialog.askstring("Master Password", "Enter Master Password to copy password to clipboard: ")
        
        # get masterpass
        with open(jPath(str(Path.home()), '.pastaman', '.m'), 'rb') as m:
            masterpassword = m.read() # this is encrypted
        
        
        t1 = threading.Thread(target=self.decryptthread, args=(masterpassword.decode('ascii'),))
        t1.start()
        t1.join()
        
        
        if allegedpass == self.den.decode('ascii'):
            t = threading.Thread(target=self.arch.decrypt, args=(self.arch.__searchresult__['password'],))
            t.start()
            t.join()
            
            pyperclip.copy(self.arch._dec_)
            spam = pyperclip.paste()
            self.arch._dec_ = None
        else:
            messagebox.showwarning("Wrong Master Password", "The master password entered by you is wrong!")
        
    def _makeAdd_(self):
        # -> create enclosing frame under Add
        self.AEF = ttk.Frame(self.addtab)
        self.AEF.pack(expand=True, fill=BOTH)
        
        # -> create target label and input
        self.targetLabel = ttk.Label(self.AEF, text='Target:')
        self.targetLabel.place(anchor='center', relx=0.24, rely=0.16)
        # --> textvar for targetEntry
        self.vartarget = StringVar()
        self.targetEntry = ttk.Entry(self.AEF, width=25, textvariable=self.vartarget)
        self.targetEntry.place(anchor='center', relx=0.57, rely=0.16)
        
        # -> create target-type label and input
        self.targtypeLabel = ttk.Label(self.AEF, text='Target Type:')
        self.targtypeLabel.place(anchor='center', relx=0.235, rely=0.28)
        # --> textvar for targtypeEntry
        self.vartargtype = StringVar()
        self.targtypeEntry = ttk.Entry(self.AEF, width=25, textvariable=self.vartargtype)
        self.targtypeEntry.place(anchor='center', rely=0.28, relx=0.57)
        
        # -> create username label and input
        self.usernameLabel = ttk.Label(self.AEF, text='Username:')
        self.usernameLabel.place(anchor='center', relx=0.235, rely=0.4)
        # --> textvar for usernameEntry
        self.varuser = StringVar()
        self.usernameEntry = ttk.Entry(self.AEF, width=25, textvariable=self.varuser)
        self.usernameEntry.place(anchor='center', rely=0.4, relx=0.57)
        
        # -> create password label and input
        self.passLabel = ttk.Label(self.AEF, text='Password:')
        self.passLabel.place(anchor='center', relx=0.235, rely=0.52)
        # --> textvar for passEntry
        self.varpass = StringVar()
        self.passEntry = ttk.Entry(self.AEF, width=25, show="-", textvariable=self.varpass)
        self.passEntry.place(anchor='center', rely=0.52, relx=0.57)
        
        # -> create Add button
        self.addbutton = ttk.Button(self.AEF, text='[ Add ]', default=ACTIVE, command=self.add)
        self.addbutton.place(anchor='center', relx=0.5, rely=0.76)
        
        # -> create status label
        self.status = StringVar()
        self.statusbar = ttk.Label(self.AEF, textvariable=self.status)
        self.statusbar.place(anchor='center', relx=0.5, rely=0.88)
        
        # -> bind pass entry with enter
        self.passEntry.bind('<Return>', self.add)
    
    def fetchstatusReset(self):
        self.fetchstatusvar.set('')
    
    def add(self, event = None):
        # create thread for add
        threading.Thread(target=self.arch.add, args=(self.targetEntry.get().strip(), self.targtypeEntry.get().strip(), self.usernameEntry.get().strip(), self.passEntry.get().strip().encode('ascii'))).start()
        
        # clean entries
        self.varpass.set('')
        self.vartarget.set('')
        self.vartargtype.set('')
        self.varuser.set('')
        
        # set status bar
        self.status.set('Added.')
        
        # reinit the _makeFetch_ to refresh newly added Entries
        for widget in self.gettab.winfo_children():
            widget.destroy()
        
        self._makeFetch_()
        
        # after 3 sec clear status bar
        self.statusbar.after(3000, self.clearstatus)
    
    def clearstatus(self):
        self.status.set('')